// test_http_get.c
// PacketLab TCP Experiment Controller Example Program
// Use pktlab measurement endpoint to issue HTTP GET request
// Should be used with pktxpmgr
//

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include "pktlab.h"
#include "pktctrl.h"

#define BUFSZ 0x10000

//
// FUNCTION DECLARATIONS
//

static bool is_good_fd(int fd);
static int create_nopen(struct pktlab_message *msg, uint8_t *laddrptr,
                        uint8_t *lportptr, uint8_t *raddrptr, uint8_t *rportptr,
                        uint8_t sktid, uint8_t family, uint8_t proto,
                        uint32_t rbufsize, const char *localip,
                        uint16_t localport, const char *ip,
                        uint16_t remoteport);
static int create_nsend(struct pktlab_message *msg, uint8_t sktid,
                        uint8_t family, uint8_t proto, uint16_t tag,
                        pktlab_time_t time, uint32_t len, const void *ptr);
static int create_nclose(struct pktlab_message *msg, uint8_t sktid);
static int try_send(struct pktctrl_obj *obj, const struct pktlab_message *msg);
static void send_n_print_err(struct pktctrl_obj *obj,
                             const struct pktlab_message *send_msg);
static void recv_msg_until_type(int type, struct pktctrl_obj *obj,
                                struct pktlab_message **recv_msg);
static void TCP_experiment(struct pktctrl_obj *obj, const char *domain, const char *path);
static void perror_exit(const char *str, int exit_val);
static void print_err_exit(int exit_val, const char *fmt, ...);

//
// FUNCTION DEFINITIONS
//

int main(int argc, char *argv[]) {
    int rst, econnfd;
    struct pktctrl_obj *endpoint_obj;
    char domain[BUFSZ] = {};
    char path[BUFSZ] = {};
    char sample_domain[] = "www.example.com";
    char sample_path[] = "/";

    if (argc == 3) {
        strncpy(path, argv[2], sizeof(path)-1);
    } else{
        strncpy(path, sample_path, sizeof(path)-1);
    }

    if (argc >= 2) {
        strncpy(domain, argv[1], sizeof(domain)-1);
    } else {
        strncpy(domain, sample_domain, sizeof(domain)-1);
    }

    fprintf(
        stderr,
        "> PacketLab Measurement Applet: HTTP GET example for '%s%s' using pktxpmgr\n",
        domain, path);

    econnfd = 0;
    assert(is_good_fd(econnfd));

    endpoint_obj = pktctrl_create_obj();
    assert(endpoint_obj != NULL);

    rst = pktctrl_raw_session(0, endpoint_obj);
    assert(rst == PKTCTRL_SUCCESS);

    TCP_experiment(endpoint_obj, domain, path);
    pktctrl_close(endpoint_obj);

    fprintf(stderr, "> Terminating\n");

    return 0;
}

bool is_good_fd(int fd) { return (fcntl(fd, F_GETFD) >= 0); }

int create_nopen(struct pktlab_message *msg, uint8_t *laddrptr,
                 uint8_t *lportptr, uint8_t *raddrptr, uint8_t *rportptr,
                 uint8_t sktid, uint8_t family, uint8_t proto,
                 uint32_t rbufsize, const char *localip, uint16_t localport,
                 const char *ip, uint16_t remoteport) {
    uint8_t transproto = proto;

    msg->type = PKTLAB_NOPEN_MESSAGE;
    msg->nopen.family = family;
    msg->nopen.sktid = sktid;
    msg->nopen.proto = proto;
    msg->nopen.rbufsz = rbufsize;

    if (transproto == PKTLAB_TCP_PROTO || transproto == PKTLAB_UDP_PROTO) {
        if (inet_pton(family == PKTLAB_IP4_PROTO ? AF_INET : AF_INET6, ip,
                      raddrptr) != 1)
            return -1;

        if (inet_pton(family == PKTLAB_IP4_PROTO ? AF_INET : AF_INET6, localip,
                      laddrptr) != 1)
            return -1;

        msg->nopen.raddrlen = (family == PKTLAB_IP4_PROTO) ? 4 : 16;
        msg->nopen.raddrptr = raddrptr;

        pktlab_set16n(rportptr, remoteport);
        msg->nopen.rportlen = 2;
        msg->nopen.rportptr = rportptr;

        msg->nopen.laddrlen = (family == PKTLAB_IP4_PROTO) ? 4 : 16;
        msg->nopen.laddrptr = laddrptr;
        pktlab_set16n(lportptr, localport);
        msg->nopen.lportlen = 2;
        msg->nopen.lportptr = lportptr;
    }

    return 0;
}

int create_nsend(struct pktlab_message *msg, uint8_t sktid, uint8_t family,
                 uint8_t proto, uint16_t tag, pktlab_time_t time, uint32_t len,
                 const void *ptr) {
    msg->type = PKTLAB_NSEND_MESSAGE;
    msg->nsend.sktid = sktid;
    msg->nsend.family = family;
    msg->nsend.proto = proto;
    msg->nsend.time = time;
    msg->nsend.tag = tag;

    switch (proto) {
    case PKTLAB_TCP_PROTO:
        msg->nsend.len = len;
        msg->nsend.ptr = ptr;
        break;
    default:
        return -1;
    }

    return 0;
}

int create_nclose(struct pktlab_message *msg, uint8_t sktid) {
    msg->type = PKTLAB_NCLOSE_MESSAGE;
    msg->nclose.sktid = sktid;
    return 0;
}

int try_send(struct pktctrl_obj *obj, const struct pktlab_message *msg) {
    int rst;
    int i;
    struct timespec time = {.tv_sec = 0, .tv_nsec = 100000000};

    rst = pktctrl_write_message(obj, msg);

    for (i = 0; i < 2 && rst == 0; ++i) {
        nanosleep(&time, NULL); // sleep 0.1 sec
        rst = pktctrl_write_message(obj, msg);
    }

    return rst;
}

void send_n_print_err(struct pktctrl_obj *obj,
                      const struct pktlab_message *send_msg) {
    if (try_send(obj, send_msg) != 1)
        perror_exit("try_send", -1);
}

void recv_msg_until_type(int type, struct pktctrl_obj *obj,
                         struct pktlab_message **recv_msg) {

    while (true) {
        if (pktctrl_read_message(obj, recv_msg) < 0)
            print_err_exit(-1, "pktctrl_read_message failed\n");

        if (recv_msg == NULL) {
            print_err_exit(-1, "pktctrl_read_message: sock EOF\n");
        } else if ((*recv_msg)->type != type) {
            fprintf(stderr, ">> Got unexpected msg: %d, expecting %d\n",
                    (*recv_msg)->type, type);
            free(*recv_msg);
            continue;
        }

        break;
    }

    return;
}

void TCP_experiment(struct pktctrl_obj *obj, const char *domain, const char *path) {
    //
    // Create and send nopen msg
    // Receive result
    //

    fprintf(stderr, "> Sending nopen msg\n");

    struct pktlab_message send_msg;
    uint8_t sktid = 1;
    uint8_t family = PKTLAB_IP4_PROTO;
    uint8_t proto = PKTLAB_TCP_PROTO;
    uint32_t rbufsize = 0x1000;
    const char localip[] = "0.0.0.0";
    uint16_t remoteport = 80;
    uint16_t localport = 0;
    uint8_t remoteaddr[PKTLAB_ADDRLEN_MAX];
    uint8_t localaddr[PKTLAB_ADDRLEN_MAX];
    uint8_t rports[2];
    uint8_t lports[2];
    char ip[INET6_ADDRSTRLEN];
    struct pktlab_message *recv_msg;
    struct addrinfo hint = {.ai_family=AF_INET};
    struct addrinfo *res;
    struct addrinfo *aiptr;
    bool addr_found;

    // Note we do the domain name resolution at the controller-side
    // One may want to do it at the endpoint side instead
    if (getaddrinfo(domain, NULL, &hint, &res) != 0)
        print_err_exit(-1, "error: getaddrinfo failed with %s\n", strerror(errno));

    addr_found = false;
    for (aiptr = res; aiptr != NULL; aiptr = aiptr->ai_next) {
        if (aiptr->ai_family != AF_INET)
            continue;

        if (inet_ntop(aiptr->ai_family,
                      &((struct sockaddr_in *) aiptr->ai_addr)->sin_addr.s_addr,
                      ip, sizeof(ip))) {
            fprintf(stderr, ">> Resolved IPv4 address: %s\n", ip);
            addr_found = true;
            break;
        }
    }

    freeaddrinfo(res);

    if (!addr_found)
        print_err_exit(-1, "error: cannot resolve domain\n");

    if (create_nopen(&send_msg, localaddr, lports, remoteaddr, rports, sktid,
                     family, proto, rbufsize, localip, localport, ip,
                     remoteport) < 0)
        perror_exit("error: create_nopen", -1);

    send_n_print_err(obj, &send_msg);
    recv_msg_until_type(PKTLAB_RESULT_MESSAGE, obj, &recv_msg);

    if (recv_msg->result.errid != PKTLAB_SUCCESS) { // only 0 imply success
        print_err_exit(-1, "result msg %d\n", (int)recv_msg->result.errid);
    }

    fprintf(stderr, ">> result msg id: %d\n", (int)recv_msg->result.errid);

    free(recv_msg);

    //
    // Create and send nsend msg
    // Receive result
    //

    fprintf(stderr, "> Sending nsend msg\n");

    char payload[BUFSZ];
    pktlab_time_t time = 0;
    uint16_t tag = 1234;
    pktlab_time_t ec_nsend_time;

    snprintf(payload, BUFSZ, "GET %s HTTP/1.0\r\nHost: %s\r\n\r\n", path, domain);

    if (create_nsend(&send_msg, sktid, family, proto, tag, time,
                     strlen(payload), payload) < 0)
        print_err_exit(-1, "create_nsend: ret -1\n");

    send_n_print_err(obj, &send_msg);
    ec_nsend_time = pktlab_time_now();
    recv_msg_until_type(PKTLAB_RESULT_MESSAGE, obj, &recv_msg);

    if (recv_msg->result.errid != 0) { // only 0 imply success
        print_err_exit(-1, "result msg %d\n", (int)recv_msg->result.errid);
    }

    fprintf(stderr, ">> result msg id: %d\n", (int)recv_msg->result.errid);

    free(recv_msg);

    unsigned long cnt = 0;
    char buf[BUFSZ];
    pktlab_time_t me_recv_time;
    pktlab_time_t ec_recv_time;
    pktlab_time_t me_send_time;
    int done = 0;
    while (!done) {
        if (pktctrl_read_message(obj, &recv_msg) < 0)
            perror_exit("pktctrl_read_message", -1);
        if (recv_msg == NULL)
            print_err_exit(-1, "pktctrl_read_message: sock EOF\n");
        fprintf(stderr, ">> Got message type: %d\n", recv_msg->type);
        switch (recv_msg->type) {
        case PKTLAB_NTAG_MESSAGE:
            assert(tag == recv_msg->ntag.tag);
            assert(recv_msg->ntag.errid == PKTLAB_SUCCESS);
            me_send_time = recv_msg->ntag.time;
            break;
        case PKTLAB_NDATA_MESSAGE:
            if (recv_msg->ndata.sktid != sktid)
                print_err_exit(-1, "Got ndata from unknown socket: %d\n",
                               recv_msg->ndata.sktid);

            if (recv_msg->ndata.len > sizeof(buf)-cnt)
                print_err_exit(-1, "Got too much ndata from endpoint\n");

            memcpy(buf + cnt, recv_msg->ndata.ptr, recv_msg->ndata.len);
            cnt += recv_msg->ndata.len;
            me_recv_time = recv_msg->ndata.time;
            ec_recv_time = pktlab_time_now();
            break;
        case PKTLAB_NSTAT_MESSAGE:
            fprintf(stderr, "nstat msg id %d\n", recv_msg->nstat.code);
            if (recv_msg->nstat.code == PKTLAB_TCPEV_PEERCLOSED) {
                done = 1;
            }
            break;
        case PKTLAB_SUSPD_MESSAGE:
            fprintf(stderr, "experiment suspended\n");
            return;
            // break;
        default:
            print_err_exit(-1, "Unknown msg type %d\n", recv_msg->type);
            break;
        }
        free(recv_msg);
    }

    //
    // Create and send nclose msg
    // Receive result
    //

    fprintf(stderr, "> Sending nclose msg\n");

    create_nclose(&send_msg, sktid);

    send_n_print_err(obj, &send_msg);
    recv_msg_until_type(PKTLAB_RESULT_MESSAGE, obj, &recv_msg);

    if (recv_msg->result.errid != 0) { // only 0 imply success
        print_err_exit(-1, "result msg %d\n", (int)recv_msg->result.errid);
    }

    free(recv_msg);

    //
    // Print rst
    //

    fprintf(stderr, "\n---->Results<----\n");
    fprintf(stderr, "Received data length: %lu\n", cnt);
    fprintf(stderr, "Spent time (in ns) at endpoint: %" PRIu64 "\n",
        me_recv_time - me_send_time);
    fprintf(stderr, "Spent time (in ns) at controller: %" PRIu64 "\n",
        ec_recv_time - ec_nsend_time);
    fprintf(stderr, "\n-->Content<--\n");
    buf[cnt] = '\0';
    fprintf(stderr, "%s\n", buf);
    exit(0);
    return;
}

void perror_exit(const char *str, int exit_val) {
    perror(str);
    exit(exit_val);
}

void print_err_exit(int exit_val, const char *fmt, ...) {
    static char buf[256];
    va_list args;

    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);
    va_end(args);

    fprintf(stderr, "%s", buf);

    exit(exit_val);
}
